""" Utilities for storage backends

Provides
--------
- Initalize an S3 bucket for ARD delivery
- Revoke ARD delivery access from an S3 bucket

Example
-------
Initialize a bucket

>>> init_bucket('my-data-bucket')

Revoke access

>>> revoke_bucket('my-data-bucket')
"""
import json
import sys

import boto3
import botocore

from max_ard.exceptions import BucketInitError, BucketRevokeError


def init_bucket(location, profile=None):
    """Set up an S3 bucket for ARD delivery

    Parameters
    ----------
    location : str
        bucket name
    profile : str, optional
        name of S3 profile to use for access

    Notes
    -----
    If the bucket does not exist, it will be created.

    This adds the required bucket policy statement to the bucket granting the ARD API
    to write objects to the bucket. No other access is granted.

    This also changes the bucket's Object Ownership policy to "Owner Preferred", which
    transfers the ownership of ARD objects to the bucket owner instead of the ARD API."""

    location = location.replace("s3://", "")
    if profile is None:
        s3_resource = boto3.resource("s3")
        s3_client = boto3.client("s3")
    else:
        session = boto3.Session(profile_name=profile)
        s3_resource = session.resource("s3")
        s3_client = session.client("s3")
    # check if the bucket exists
    try:
        s3_resource.meta.client.head_bucket(Bucket=location)
        print(f"Bucket exists at {location}")
    except botocore.exceptions.ClientError as e:
        if "403" in str(e):
            msg = "That bucket name already exists in AWS. If you made the bucket, your user does not have permission to see it"
            raise BucketInitError(msg)
        try:
            s3_resource.create_bucket(Bucket=location)
            print(f'Created bucket "{location}"')
        except botocore.exceptions.ClientError as e:
            if "ExpiredToken" in str(e):
                msg = "Your AWS token has expired, please request a new token"
                raise BucketInitError(msg)
            else:
                raise e
        except botocore.errorfactory.BucketAlreadyExists:
            msg = "That bucket name already exists in AWS, try a different name"
            raise BucketInitError(msg)
    authorized = False
    try:
        policy = json.loads(s3_client.get_bucket_policy(Bucket=location)["Policy"])
    except botocore.exceptions.ClientError:
        policy = {"Version": "2012-10-17", "Statement": []}
    for statement in policy["Statement"]:
        if statement["Sid"] == "MaxarArdS3Access":
            authorized = True
            break
    if not authorized:
        statement = {
            "Sid": "MaxarARDS3Access",
            "Effect": "Allow",
            "Principal": {"AWS": "arn:aws:iam::334489843805:root"},
            "Action": "s3:PutObject",
            "Resource": [f"arn:aws:s3:::{location}/*", f"arn:aws:s3:::{location}"],
        }
        policy["Statement"].append(statement)
        s3_client.put_bucket_policy(Bucket=location, Policy=json.dumps(policy))
        print("Added ARD writer policy to bucket")
    else:
        print("Bucket already authorized")
    response = s3_client.put_bucket_ownership_controls(
        Bucket=location,
        OwnershipControls={
            "Rules": [
                {"ObjectOwnership": "BucketOwnerPreferred"},
            ]
        },
    )
    print("Object Ownership set to Bucket Owner Preferred")


def revoke_bucket(location, profile=None):
    """Revoke ARD API write access to a bucket

    Parameters
    ----------
    location : str
        bucket name
    profile : str, optional
        name of S3 profile to use for access

    Notes
    -----
    This removes the required bucket policy statement from the bucket granting the ARD API
    to write objects to the bucket.

    This does not change the bucket's Object Ownership policy."""

    location = location.replace("s3://", "")
    if profile is None:
        s3_resource = boto3.resource("s3")
        s3_client = boto3.client("s3")
    else:
        session = boto3.Session(profile_name=profile)
        s3_resource = session.resource("s3")
        s3_client = session.client("s3")

    # check if the bucket exists
    try:
        s3_resource.meta.client.head_bucket(Bucket=location)
    except botocore.exceptions.ClientError:
        msg = "Either this bucket does not exist or you do not have permission to access it"
        raise BucketRevokeError(msg)
    try:
        policy = json.loads(s3_client.get_bucket_policy(Bucket=location)["Policy"])
    except botocore.exceptions.ClientError:
        msg = "Bucket has no policy statements - nothing to revoke"
        raise BucketRevokeError(msg)
    statements = []
    for statement in policy["Statement"]:
        if statement["Sid"] != "MaxarARDS3Access":
            statements.append(statement)
    if len(statements) < len(policy["Statement"]):
        if len(statements) == 0:
            s3_client.delete_bucket_policy(Bucket=location)
        else:
            policy["Statement"] = statements
            s3_client.put_bucket_policy(Bucket=location, Policy=json.dumps(policy))
        print("Removed ARD writer policy from bucket")
    else:
        msg = "Could not find ARD writer policy in bucket"
        raise BucketRevokeError(msg)



""" Convert ARDCollections to QGIS Layer format"""

import warnings

warnings.warn(
    "max_ard.outputs.qlr_format.QlrDoc has been moved to max_ard.io.QlrDoc",
    DeprecationWarning,
    stacklevel=2,
)

from max_ard.io import QlrDoc as NewQlrDoc


def QlrDoc(collection, public=False):
    """Return a an ARDCollection as a QGIS layer file

    Parameters
    ----------
    collection : ARDCollection
    public : bool, optional
      True if the data is in a public S3 bucket location (see Notes)

    Returns
    -------
    str
      QLR document contents of the ARDCollection

    Notes
    -----
    Passing `public=true` will convert s3:// URLs to the public http:// S3 endpoints.
    This functionality does not work yet for other storage types. It's functionality
    is primarily for creating QLRs of the official Maxar ARD sample datasets at
    http://maxar-ard-samples.s3-website-us-east-1.amazonaws.com/
    """

    warnings.warn(
        "max_ard.outputs.qlr_format.QlrDoc has been moved to max_ard.io.QlrDoc",
        DeprecationWarning,
        stacklevel=2,
    )
    return NewQlrDoc(collection, public)



"""Convert ARD Selects to a Shapefile"""

import warnings

from max_ard.io import ShpDoc as NewShpDoc


def ShpDoc(geojson, path):
    """Convert a Select represented as a geojson to a shapefile and save

    Parameters
    ----------
    geojson: geojson representing a Select
    path: location to save the shapefile

    Returns
    -------
    none"""

    warnings.warn(
        "max_ard.outputs.shp_format.ShpDoc has been moved to max_ard.io.ShpDoc",
        DeprecationWarning,
        stacklevel=2,
    )
    return NewShpDoc(geojson, path)



"""Convert ARD Selects to a Shapefile"""

__all__ = ("ShpDoc",)

import json

from shapely.geometry import Polygon, mapping

from max_ard.dependency_support import HAS_FIONA
from max_ard.exceptions import MissingDependency

if HAS_FIONA:
    import fiona
    from fiona.crs import from_epsg


def ShpDoc(geojson, path):
    """Convert a Select represented as a geojson to a shapefile and save

    Parameters
    ----------
    geojson: geojson representing a Select
    path: location to save the shapefile

    Returns
    -------
    none"""
    if not HAS_FIONA:
        raise MissingDependency("Shapefile export requires Fiona")

    dict = json.loads(geojson)

    props = {
        "cell_id": "str",
        "available_tiles": "int",
        "stack_depth_fulfilled": "bool",
        "acquisition_datetime_min": "str",
        "acquisition_datetime_max": "str",
        "best_matches": "str",
    }

    schema = {
        "geometry": "Polygon",
        "properties": props,
    }

    # Write a new shapefile
    with fiona.open(path, "w", crs=from_epsg(4326), driver="ESRI Shapefile", schema=schema) as shp:

        for feat in dict["features"]:

            tuple_coords = [tuple(coords) for coords in feat["geometry"]["coordinates"][0]]
            polygon = Polygon(tuple_coords)
            p = feat["properties"]
            best_matches = p["best_matches"]

            best_str = ""
            for i in range(len(best_matches)):
                best_str += str(best_matches[i])
                if i < len(best_matches) - 1:
                    best_str += ", "

            p["best_matches"] = best_str

            shp.write({"geometry": mapping(polygon), "properties": p})



import warnings

warnings.warn(
    "max_ard.outputs.kml_format.KmlDoc has been moved to max_ard.io.KmlDoc", DeprecationWarning
)

from max_ard.io import KmlDoc as NewKmlDoc


def KmlDoc(select):
    """Convert a Select to a KML file

    Parameters
    ----------
    select: Select

    Returns
    -------
    str
        A KML document of the Select"""

    warnings.warn(
        "max_ard.outputs.kml_format.KmlDoc has been moved to max_ard.io.KmlDoc", DeprecationWarning
    )
    return NewKmlDoc(select)



"""Base collection types for ARD

Provides
--------
- Pydantic Models that are used in multiple places
- Submitted and Succeeded ABCs that add "requires" decorators
- Collections of tiles
- Collections of tiles from a common acquisition
- Collections of tiles covering a common grid cells

These collection objects assume a generic concept of a "tile". They can be used to
group tiles-to-be found in a Select, or real tiles delivered to a storage location.
"""
from abc import ABC, abstractmethod
from functools import wraps
from tempfile import NamedTemporaryFile
from typing import Any, List, Union

from maxar_ard_grid.grid import Cell
from pydantic import BaseModel
from shapely.geometry import MultiPolygon, mapping

from max_ard.dependency_support import HAS_RASTERIO
from max_ard.exceptions import NotFinished, NotSubmitted, NotSuccessful, SelectError
from max_ard.processing import AcquisitionReader

if HAS_RASTERIO:
    import rasterio


def hydrate(cls, things, key=None):
    """Convenience function for hydrating objects from a list of dicts

    Arguments
    ---------
    cls: Object class to create
    things: iterable of raw object dicts
    key: like in sort(), function that exposes where in the dict the object's arguments are

    Returns
    -------
    list
        list of objects of type cls

    Notes
    -----
    If a key function is not provided, the dict is assumed to hold the object arguments.
    Operator.itemgetter(<name>) or a lamba x: x[<name>] can be used.

    """
    if key is None:
        key = lambda x: x
    return [cls(**key(thing)) for thing in things]


def hydrate_with_responses(cls, response_cls, responses, **kwargs):
    instances = []
    for response in responses:
        instance = cls(**kwargs)
        instance.response = response_cls(**response)
        instances.append(instance)

    return instances


###
# Common Pydantic Models
###


class ARDModel(BaseModel):
    """A Pydantic BaseModel with sanitized dicts

    The ARD API can be picky about Nones and empty objects so for request payloads
    we can use the to_payload method to generate a cleaned up payload"""

    def to_payload(self):
        """sends only parameters with values"""
        params = self.dict().copy()
        payload = {k: v for k, v in params.items() if v not in (None, (), [])}
        return payload


class EmailNotification(BaseModel):
    """A model for Email Notifications"""

    type: str = "email"
    address: str

    class Config:
        allow_mutation: False


class SNSNotification(BaseModel):
    """A model for SNS Notifications"""

    type: str = "sns"
    topic_arn: str

    class Config:
        allow_mutation: False


class UsageArea(BaseModel):
    """A model to hold ARD usage by area"""

    fresh_imagery_sqkm: float
    standard_imagery_sqkm: float
    training_imagery_sqkm: float
    tasking_imagery_sqkm: Any
    total_imagery_sqkm: float
    estimate: bool


class UsageCost(BaseModel):
    """A model to hold ARD image costs, estimated or consumed

    A value of None for a cost means your account does not have pricing set
    for this category. Without set pricing this category can not be ordered."""

    fresh_imagery_cost: Union[float, None]
    standard_imagery_cost: Union[float, None]
    training_imagery_cost: Union[float, None]
    tasking_imagery_cost: Any
    total_imagery_cost: Union[float, None]
    estimate: bool


class UsageLimits(BaseModel):
    """A model to hold account limits"""

    limit_sqkm: float
    fresh_imagery_fee_limit: float
    standard_imagery_fee_limit: float
    training_imagery_fee_limit: float
    tasking_imagery_fee_limit: Any
    annual_subscription_fee_limit: float


class UsageAvailable(BaseModel):
    """A model to hold account balances

    When returned with a select, these numbers reflect the account balance at the time
    the select was submitted. Usage estimates are not applied to these numbers."""

    available_sqkm: float
    fresh_imagery_balance: float
    standard_imagery_balance: float
    training_imagery_balance: float
    tasking_imagery_balance: Any
    total_imagery_balance: float


###
# Validation ABCs
###


class Submitted(ABC):
    """An ABC for objects that can be submitted

    - Concrete classes that inherit from Submitted must override submitted()
    - `submitted` is usually exposed as a @property
    - how you determine an object has been submitted is up to you

    For methods that require the object to have been submitted, wrap it with
    the @Submitted.required decorator
    """

    @abstractmethod
    def submitted(self):
        pass

    @staticmethod
    def required(func):
        @wraps(func)
        def func_wrapper(instance, *args, **kwargs):
            if not instance.submitted:
                raise NotSubmitted(
                    f"Can't access `{func.__name__}` until {type(instance).__name__} has been submitted"
                )
            res = func(instance, *args, **kwargs)
            return res

        return func_wrapper


class Succeeded(ABC):
    """An ABC for objects that have a succeeded state, and by definition
    must also be submitted in order to run

    - Concrete classes that inherit from Submitted must override
        - submitted()
        - finished()
        - succeeded()
    - see also Submitted

    For methods that require the object to have succeeded, wrap it with
    the @Succeeded.required decorator
    """

    @abstractmethod
    def submitted(self):
        pass

    @abstractmethod
    def finished(self):
        pass

    @abstractmethod
    def succeeded(self):
        pass

    @staticmethod
    def required(func):
        @wraps(func)
        def func_wrapper(instance, *args, **kwargs):
            if not instance.submitted:
                raise NotSubmitted(
                    f"Can't access `{func.__name__}` until {type(instance).__name__} has been submitted"
                )
            if not instance.finished:
                raise NotFinished(
                    f"Can't access `{func.__name__}` until {type(instance).__name__} has finished running"
                )
            if not instance.succeeded:
                # This error should probably be generic, but only Selects use this ABC so far
                if type(instance).__name__ == "Select":
                    raise SelectError(f"Can't access `{func.__name__}` due to an error")
                else:
                    raise NotSuccessful(f"Can't access `{func.__name__}` due to an error")

            res = func(instance, *args, **kwargs)
            return res

        return func_wrapper


###
# Classes for ARD Concepts
###


class Acquisition(list):
    """Container for one or more tiles from a single acquisition

    These are generated by ARDCollections and would not normally be initialized"""

    def __init__(self, acq_id: str, tiles: list = []) -> None:

        self.acq_id = acq_id
        """ str: ID of this acquisition """
        super().__init__(*tiles)

    @property
    def cells(self) -> List[Cell]:
        cells = set([tile.cell for tile in self])
        return list(cells)

    def get_tile_from_cell(self, cell):
        cell = Cell(cell)
        for tile in self:
            if tile.cell == cell:
                return tile
        raise ValueError("Tile not found in this acquisition")

    def __str__(self):
        return f"<Acquisition at {self.acq_id} ({len(self)} tiles)>"

    def __repr__(self):
        tiles = ", ".join([f"<{type(tile).__name__} at {tile.cell.id}>" for tile in self])
        return f"<Acquisition of {self.acq_id} [{tiles}]>"

    def __hash__(self):
        """Hash of cell ID"""
        return hash(self.acq_id)

    def open_acquisition(self):
        """Return a Rasterio reader that accesses all of the Acquisition object's tiles.

        A Rasterio DatasetReader opened on a VRT of all acquisition tiles"""

        return AcquisitionReader(self)

    @property
    def date(self):
        return self[0].date

    @property
    def properties(self) -> dict:
        """Strip level properties"""

        props = self[0].properties
        # todo: check if catalog_id is still passed in metadata
        keep = ["catalog_id", "acquisition_id", "platform", "scan_direction", "epsg", "datetime"]
        strip_props = {}
        for key in keep:
            try:
                strip_props[key] = props[key]
            except KeyError:
                pass
        return strip_props

    @property
    def __geo_interface__(self) -> dict:
        geoms = [tile.cell.geom_WGS84 for tile in self]
        return mapping(MultiPolygon(geoms))


class Stack(list):
    """Container for one or more tiles in an ARD grid cell

    These are generated by ARDCollections and would not normally be initialized"""

    def __init__(self, cell: Cell, tiles: list = []) -> None:

        self.cell = Cell(cell)
        super().__init__(*tiles)

    @property
    def tiles(self):
        """not needed but provides access to tiles for GeoMixin"""
        return self

    def get_tile_from_acquisition(self, acq_id):
        for tile in self:
            if tile.acq_id == acq_id:
                return tile
        raise ValueError("Tile not found in this stack")

    @property
    def acquisition_ids(self) -> List[str]:
        ids = set([tile.acq_id for tile in self])
        return list(ids)

    def __str__(self):
        return f"<Stack at {self.cell.id} ({len(self)} tiles)>"

    def __repr__(self):
        tiles = ", ".join([f"<{type(tile).__name__} of {tile.acq_id}>" for tile in self])
        return f"<Stack at {self.cell.id} [{tiles}]>"

    def __hash__(self):
        """Hash of cell ID"""
        return hash(self.cell)

    @property
    def __geo_interface__(self) -> dict:
        geoms = [tile.cell.geom_WGS84 for tile in self]
        return mapping(MultiPolygon(geoms))


class Store(dict):
    """A defaultdict-like k:v store for objects"""

    def __init__(self, object_type, *args, **kwargs):
        self.object_type = object_type
        super().__init__(*args, **kwargs)

    def __missing__(self, key):
        res = self[key] = self.object_type(key)
        return res


class BaseCollection:
    def __init__(self):
        self._reset()

    def _reset(self):
        """Used by ARDCollection when a rescan is triggered"""
        self._acquisitions = Store(Acquisition)
        self._stacks = Store(Stack)
        self._zones = set()
        self._dates = set()

    def add_tile(self, tile):
        self._acquisitions[tile.acq_id].append(tile)
        self._stacks[tile.cell].append(tile)
        self._zones.add(tile.cell.zone)
        self._dates.add(tile.date)

    @property
    def tiles(self):
        """All the ARD Tiles in this collection"""
        return [tile for tiles in self.stacks for tile in tiles]

    def get_tile(self, acq_id, cell):
        """Get a tile from the collection

        Parameters
        ----------
        acq_id : str
            Acquisition ID
        cell : str or maxar_ard_grid.Cell
            Cell of the tile, can be a cell ID or Cell object

        Returns
        -------
        ARDTile or SelectTile
        """
        cell = Cell(cell)
        acq = self.get_acquisition(acq_id)
        for tile in acq:
            if tile.cell == cell:
                return tile
        raise ValueError("Tile not found in this collection")

    @property
    def dates(self) -> List[str]:
        """All of the dates of acquisitions in the ARDCollection"""
        dates = list(self._dates)
        dates.sort()
        return dates

    @property
    def start_date(self) -> str:
        """Earliest date of acquisitions in the ARDCollection as YYYY-MM-DD"""
        return self.dates[0]

    @property
    def end_date(self) -> str:
        # does the type get caught?
        """Latest date of acquisitions in the ARDCollection as YYYY-MM-DD"""
        return self.dates[-1]

    @property
    def zones(self) -> List[int]:
        """The UTM zones covered by tiles in the ARDCollection"""
        return list(self._zones)

    #
    # Stacks
    #

    @property
    def stacks(self):
        """All of ARD grid stacks in the collection"""
        return list(self._stacks.values())

    @property
    def cells(self):
        """All of the ARD grid Cell objects in the collection"""
        return list(self._stacks.keys())

    def get_stack(self, cell):
        """Get a stack of tiles from the collection

        A Stack object is a container of all tiles covering a given ARD grid cell.
        Parameters
        ----------
        cell : str or maxar_ard_grid.Cell
            Cell of the tile, can be a cell ID or Cell object

        Returns
        -------
        Stack
        """
        cell = Cell(cell)
        try:
            return self._stacks[cell]
        except KeyError:
            raise ValueError(f"Stack for cell {cell} not found in this collection")

    #
    # Acquisitions
    #

    @property
    def acquisitions(self):
        """All of the acquisitions present in the collection"""
        return list(self._acquisitions.values())

    @property
    def acquisition_ids(self):
        """All of the acquisitions IDs present in the collection"""
        return list(self._acquisitions.keys())

    @property
    def acq_ids(self):
        """Shorthand name for .acquisition_ids"""
        return self.acquisition_ids

    def get_acquisition(self, acq_id):
        """Get an acquisition from the collection

        An Acquisition object is a container of all tiles from a given acquisition.

        Parameters
        ----------
        acq_id : str
            Acquisition ID

        Returns
        -------
        Acquisition
        """
        try:
            return self._acquisitions[acq_id]
        except KeyError:
            raise ValueError(f"Aquisition for ID {acq_id} not found in this collection")

    @property
    def __geo_interface__(self) -> dict:
        geoms = [tile.cell.geom_WGS84 for tile in self.tiles]
        return mapping(MultiPolygon(geoms))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['max_ard', 'max_ard.commands', 'max_ard.io', 'max_ard.outputs']

package_data = \
{'': ['*']}

install_requires = \
['Shapely>=1.7.1,<2.0.0',
 'backports.cached-property>=1.0.1,<2.0.0',
 'boto3>=1.17.73,<2.0.0',
 'click>=8.0.0,<9.0.0',
 'maxar-ard-grid>=1.2.1,<2.0.0',
 'pdoc3>=0.10.0,<0.11.0',
 'pydantic>=1.7.4,<2.0.0',
 'requests-futures>=1.0.0,<2.0.0',
 'requests-oauthlib>=1.3.0,<2.0.0',
 'requests>=2.25.1,<3.0.0',
 's3fs>=0.4.2,<0.5.0']

extras_require = \
{'full': ['rasterio>=1.2.10,<2.0.0', 'Fiona>=1.8.19,<2.0.0']}

entry_points = \
{'console_scripts': ['max-ard = max_ard.commands.command:max_ard']}

setup_kwargs = {
    'name': 'max-ard',
    'version': '1.6.0',
    'description': 'ARD SDK and CLI tools',
    'long_description': '# max_ard\n\nA Python SDK and CLI for working with [Maxar ARD Imagery](https://ard.maxar.com/docs/)\n\n## Introduction\n\n`max_ard` consists of a Python SDK for selecting, ordering, and working with Maxar ARD imagery and a set of CLI tools for common tasks.\n\n## Installation\n\n`pip install max-ard`\n\n## Documentation\n\nExtensive docs are available at https://ard.maxar.com/docs/sdk/\n',
    'author': 'Marc Pfister',
    'author_email': 'marc.pfister@maxar.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://ard.maxar.com/docs/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7.0,<4.0.0',
}


setup(**setup_kwargs)
